// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2017 and onwards Google, Inc.

#ifndef BAUMWELCH_DECODE_H_
#define BAUMWELCH_DECODE_H_

// This header defines functions for decoding ciphertext into plaintext using
// the Viterbi algorithm. Both FST and FAR inputs/outputs are supported.
//
// If the semiring (arc) does not have the path property, then there must exist
// a weight converter to and from the tropical semiring (arc). (Weight
// properties must also be known at compile time.)
//
// Decoding proceeds as follows:
//
// 1. Compose the channel model and the ciphertext.
// 2. Project onto the input (plaintext) side of (1).
// 3. Compose the plaintext with the result of (2) to construct a weighted
//    lattice.
// 4. If the input semiring is does not have the path property (i.e., is not
//    kPath), attempt to convert the weighted lattice to the tropical semiring.
// 5. Compute the shortest path over the weighted lattice (3-4).
// 6. If the input semiring does not have the path property (i.e., is not
//    kPath), attempt to convert the shortest path (5) back to the input
//    semiring.
// 7. Project onto the input, remove epsilon arcs, and remove weights.
//
// Note that if a uniform channel model is provided, this is equivalent to
// decoding with the plaintext model alone.
//
// A user-facing function SparsityBias is provided. When applied to a channel
// with a power argument > 1, a one-time sparsity bias will be imposed. (A
// cubic penalty is a good place to start.) This only makes sense when applied
// to a trained channel model.
//
// By convention, operations in this library that work with FarReader input
// reset the FAR to its initial position upon completion.

#include <type_traits>

#include <fst/extensions/far/far.h>
#include <fst/fstlib.h>

namespace fst {
namespace internal {

// Power-mapper that takes a scalar power argument. This template is only valid
// for unsigned float weight types.

template <class A>
class FloatPowerMapper {
 public:
  using FromArc = A;
  using ToArc = A;
  using Weight = typename FromArc::Weight;
  using ValueType = typename Weight::ValueType;

  explicit FloatPowerMapper(ValueType power) : power_(power) {}

  ToArc operator()(const FromArc &arc) const {
    return ToArc(arc.ilabel, arc.olabel, Power(arc.weight, power_),
                 arc.nextstate);
  }

  constexpr MapFinalAction FinalAction() const { return MAP_NO_SUPERFINAL; }

  constexpr MapSymbolsAction InputSymbolsAction() const {
    return MAP_COPY_SYMBOLS;
  }

  constexpr MapSymbolsAction OutputSymbolsAction() const {
    return MAP_COPY_SYMBOLS;
  }

  uint64 Properties(uint64 props) const {
    return props & kWeightInvariantProperties;
  }

 private:
  const ValueType power_;
};

// Performs the shortest-path operation in the tropical semiring then maps
// back to the input semiring.
//
// This operation is only valid if there exists a valdi weight conversion
// functor for converting from Arc::Weight to TropicalArc (as defined below).
//
// This operation is unnecessary if FST::Arc::Weight is itself a kPath semiring.
template <class Arc, class FST>
void TropicalShortestPath(const Fst<Arc> &lattice, FST *plaintext) {
  static_assert(std::is_same<Arc, typename FST::Arc>::value,
                "Arc types do not match");
  using TropicalArc =
      ArcTpl<TropicalWeightTpl<typename Arc::Weight::ValueType>>;
  using ToMapper = WeightConvertMapper<Arc, TropicalArc>;
  ArcMapFst<Arc, TropicalArc, ToMapper> mapped_lattice(lattice, ToMapper());
  VectorFst<TropicalArc> shortest_path;
  ShortestPath(mapped_lattice, &shortest_path);
  ArcMap(shortest_path, plaintext, WeightConvertMapper<TropicalArc, Arc>());
}

// Computes the shortest path (over a semiring with the path property).
template <class Arc,
          typename std::enable_if<(Arc::Weight::Properties() & kPath) ==
                                  kPath>::type * = nullptr>
void ComputeShortestPath(const Fst<Arc> &ico, MutableFst<Arc> *plaintext) {
  ShortestPath(ico, plaintext);
}

// Computes the shortest path (over a semiring without the path property).
template <class Arc,
          typename std::enable_if<(Arc::Weight::Properties() & kPath) !=
                                  kPath>::type * = nullptr>
void ComputeShortestPath(const Fst<Arc> &ico, MutableFst<Arc> *plaintext) {
  TropicalShortestPath(ico, plaintext);
}

// Projects, removes epsilons, and removes weights, producing an unweighted
// string FSA.
template <class Arc>
void Tidy(MutableFst<Arc> *plaintext) {
  Project(plaintext, PROJECT_INPUT);
  RmEpsilon(plaintext);
  ArcMap(plaintext, RmWeightMapper<Arc>());
  TopSort(plaintext);  // Required by kString due to weird state ordering.
}

}  // namespace internal

// Decodes a pair of plaintext/ciphertext FARs into a FAR writer (pair
// construction).
template <class Arc>
void DecodeBaumWelch(FarReader<Arc> *plaintext, FarReader<Arc> *ciphertext,
                     const Fst<Arc> &channel, FarWriter<Arc> *hypotext) {
  // In the case they both do not contain the same number of FARs, we terminate
  // early.
  for (; !(plaintext->Done() || ciphertext->Done());
       plaintext->Next(), ciphertext->Next()) {
    VectorFst<Arc> hypotext_fst;
    const ComposeFst<Arc> co(channel, *ciphertext->GetFst());
    const ComposeFst<Arc> ico(*plaintext->GetFst(), co);
    internal::ComputeShortestPath(ico, &hypotext_fst);
    hypotext->Add(plaintext->GetKey() + "_" + ciphertext->GetKey(),
                  hypotext_fst);
  }
  plaintext->Reset();
  ciphertext->Reset();
}

// Decodes a ciphertext FAR into a FAR writer (decipherment construction).
// The output is string-compacted.
template <class Arc>
void DecodeBaumWelch(const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext,
                     const Fst<Arc> &channel, FarWriter<Arc> *hypotext) {
  for (; !ciphertext->Done(); ciphertext->Next()) {
    VectorFst<Arc> hypotext_fst;
    const ComposeFst<Arc> co(channel, *ciphertext->GetFst());
    const ComposeFst<Arc> ico(plaintext, co);
    internal::ComputeShortestPath(ico, &hypotext_fst);
    internal::Tidy(&hypotext_fst);
    hypotext->Add(ciphertext->GetKey(), CompactStringFst<Arc>(hypotext_fst));
  }
  ciphertext->Reset();
}

// Same, but with a sparsity penalty.
template <class Arc>
void DecodeBaumWelch(const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext,
                     const Fst<Arc> &channel,
                     typename Arc::Weight::ValueType power,
                     FarWriter<Arc> *hypotext) {
  using FPMapper = internal::FloatPowerMapper<Arc>;
  ArcMapFst<Arc, Arc, FPMapper> sparse_channel(channel, FPMapper(power));
  DecodeBaumWelch(plaintext, ciphertext, sparse_channel, hypotext);
}

}  // namespace fst

#endif  // BAUMWELCH_DECODE_H_

